@@ -1,6 +1,6 @@
 ;;; treemacs.el --- A tree style file viewer package -*- lexical-binding: t -*-
 
-;; Copyright (C) 2020 Alexander Miller
+;; Copyright (C) 2021 Alexander Miller
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -16,36 +16,49 @@
 ;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
-;;; Code for dealing with asynchronous processes.
+
+;; Code for dealing with treemacs' asynchronous features.
 
 ;;; Code:
 
 (require 'dash)
 (require 'ht)
 (require 's)
+(require 'vc-hooks)
 (require 'pfuture)
 (require 'treemacs-core-utils)
 (require 'treemacs-customization)
 (require 'treemacs-workspaces)
 (require 'treemacs-dom)
-(eval-and-compile
+(require 'treemacs-logging)
+(require 'treemacs-visuals)
+
+(eval-when-compile
   (require 'inline)
   (require 'treemacs-macros))
 
-(defconst treemacs--dirs-to-collpase.py
+(treemacs-import-functions-from treemacs-rendering
+  treemacs-do-delete-single-node)
+
+(defconst treemacs--dirs-to-collapse.py
   (if (member "treemacs-dirs-to-collapse.py" (directory-files treemacs-dir))
-      (f-join treemacs-dir "treemacs-dirs-to-collapse.py")
-    (f-join treemacs-dir "src/scripts/treemacs-dirs-to-collapse.py")))
+      (treemacs-join-path treemacs-dir "treemacs-dirs-to-collapse.py")
+    (treemacs-join-path treemacs-dir "src/scripts/treemacs-dirs-to-collapse.py")))
 
 (defconst treemacs--git-status.py
   (if (member "treemacs-git-status.py" (directory-files treemacs-dir))
-      (f-join treemacs-dir "treemacs-git-status.py")
-    (f-join treemacs-dir "src/scripts/treemacs-git-status.py")))
+      (treemacs-join-path treemacs-dir "treemacs-git-status.py")
+    (treemacs-join-path treemacs-dir "src/scripts/treemacs-git-status.py")))
 
 (defconst treemacs--single-file-git-status.py
   (if (member "treemacs-single-file-git-status.py" (directory-files treemacs-dir))
-      (f-join treemacs-dir "treemacs-single-file-git-status.py")
-    (f-join treemacs-dir "src/scripts/treemacs-single-file-git-status.py")))
+      (treemacs-join-path treemacs-dir "treemacs-single-file-git-status.py")
+    (treemacs-join-path treemacs-dir "src/scripts/treemacs-single-file-git-status.py")))
+
+(defconst treemacs--find-ignored-files.py
+  (if (member "treemacs-find-ignored-files.py" (directory-files treemacs-dir))
+      (treemacs-join-path treemacs-dir "treemacs-find-ignored-files.py")
+    (treemacs-join-path treemacs-dir "src/scripts/treemacs-find-ignored-files.py")))
 
 (defvar treemacs--git-cache-max-size 60
   "Maximum size for `treemacs--git-cache'.
@@ -55,9 +68,9 @@ If it does reach that size it will be cut back to 30 entries.")
   "Stores the results of previous git status calls for directories.
 Its effective type is HashMap<FilePath, HashMap<FilePath, Char>>.
 
-These cached results are used as a standin during immediate rendering when
-`treemacs-git-mode' is set to be deferred, so as to minimize the effect of large
-face changes, epsecially when a full project is refreshed.
+These cached results are used as a stand-in during immediate rendering when
+`treemacs-git-mode' is set to be deferred, so as to minimise the effect of large
+face changes, especially when a full project is refreshed.
 
 Since this table is a global value that can effectively grow indefinitely its
 value is limited by `treemacs--git-cache-max-size'.")
@@ -80,12 +93,16 @@ DEFAULT: Face"
        ("R" 'treemacs-git-renamed-face)
        (_   ,default)))))
 
+(defvar treemacs--git-mode nil
+  "Saves the specific version of git-mode that is active.
+Values are either `simple', `extended', `deferred' or nil.")
+
 (define-inline treemacs--get-node-face (path git-info default)
   "Return the appropriate face for PATH based on GIT-INFO.
 If there is no git entry for PATH return DEFAULT.
 
 PATH: Filepath
-GIT-INFO: Hashtable
+GIT-INFO: Hash-Table
 DEFAULT: Face"
   (declare (pure t) (side-effect-free t))
   (inline-letevals (path git-info default)
@@ -146,16 +163,13 @@ Real implementation will be `fset' based on `treemacs-git-mode' value."
 
 (defun treemacs--parse-git-status-extended (git-future)
   "Parse the git status derived from the output of GIT-FUTURE.
-The real parsing and formatting is done by the python process. All that's really
-left to do is pick up the cons list and put it in a hash table.
+The real parsing and formatting is done by the python process.  All that's
+really left to do is pick up the cons list and put it in a hash table.
 
 GIT-FUTURE: Pfuture"
   (or (when git-future
         (let* ((git-output (pfuture-await-to-finish git-future))
-               ;; Check fboundp in case an old version of pfuture is used.
-               ;; TODO: Remove the check when pfuture 1.7 has been widely adopted.
-               (git-stderr (when (fboundp 'pfuture-stderr)
-                             (pfuture-stderr git-future))))
+               (git-stderr (pfuture-stderr git-future)))
           ;; Check stderr separately from parsing, often git status displays
           ;; warnings which do not affect the final result.
           (unless (s-blank? git-stderr)
@@ -164,8 +178,8 @@ GIT-FUTURE: Pfuture"
                                       (s-truncate 80 it)
                                       (propertize it 'face 'error))))
               (if (< (length git-stderr) 80)
-                  (treemacs-log "treemacs-git-status.py wrote to stderr: %s" visible-error)
-                (treemacs-log "treemacs-git-status.py wrote to stderr (see full output in *Messages*): %s" visible-error)
+                  (treemacs-log-err "treemacs-git-status.py wrote to stderr: %s" visible-error)
+                (treemacs-log-err "treemacs-git-status.py wrote to stderr (see full output in *Messages*): %s" visible-error)
                 (let ((inhibit-message t))
                   (treemacs-log "treemacs-git-status.py wrote to stderr: %s" git-stderr)))))
           (when (= 0 (process-exit-status git-future))
@@ -173,14 +187,14 @@ GIT-FUTURE: Pfuture"
               (if (hash-table-p parsed-output)
                   parsed-output
                 (let ((inhibit-message t))
-                  (treemacs-log "treemacs-git-status.py output: %s" git-output))
-                (treemacs-log "treemacs-git-status.py did not output a valid hash table. See full output in *Messages*.")
+                  (treemacs-log-err "treemacs-git-status.py output: %s" git-output))
+                (treemacs-log-err "treemacs-git-status.py did not output a valid hash table. See full output in *Messages*.")
                 nil)))))
       (ht)))
 
 (defun treemacs--git-status-process-simple (path)
   "Start a simple git status process for files under PATH."
-  (let* ((default-directory (f-canonical path))
+  (let* ((default-directory (file-truename path))
          (process-environment (cons "GIT_OPTIONAL_LOCKS=0" process-environment))
          (future (pfuture-new "git" "status" "--porcelain" "--ignored" "-z" ".")))
     (process-put future 'default-directory default-directory)
@@ -215,7 +229,7 @@ GIT-FUTURE: Pfuture"
                       (if (eq ?R (aref status 0))
                           (setq i (1+ i))
                         (ht-set! git-info-hash
-                                 (f-join git-root (s-trim-left path))
+                                 (treemacs-join-path git-root (s-trim-left path))
                                  (substring (s-trim-left status) 0 1)))))
                   (setq i (1+ i)))))))))
     git-info-hash))
@@ -223,7 +237,7 @@ GIT-FUTURE: Pfuture"
 ;; TODO(2019/11/06): re-get git status when btn is flattened
 (defun treemacs--apply-deferred-git-state (parent-btn git-future buffer)
   "Apply the git fontification for direct children of PARENT-BTN.
-GIT-FUTURE is parsed the same way as in `treemacs--create-branch'. Additionally
+GIT-FUTURE is parsed the same way as in `treemacs--create-branch'.  Additionally
 since this function is run on an idle timer the BUFFER to work on must be passed
 as well since the user may since select a different buffer, window or frame.
 
@@ -259,24 +273,25 @@ BUFFER: Buffer"
   "Update the FILE node's git state, wrapped in `treemacs-save-position'.
 Internally calls `treemacs-do-update-single-file-git-state'.
 
-FILE: Filepath"
+FILE: FilePath"
   (treemacs-save-position
    (treemacs-do-update-single-file-git-state file)))
 
 (defun treemacs-do-update-single-file-git-state (file &optional exclude-parents override-status)
   "Asynchronously update the given FILE node's git fontification.
 Since an update to a single node can potentially also mean a change to the
-states of all its parents they will likewise be updated by this function. If the
-file's current and new git status are the same this function will do nothing.
+states of all its parents they will likewise be updated by this function.  If
+the file's current and new git status are the same this function will do
+nothing.
+
+When EXCLUDE-PARENTS is non-nil only the given FILE only the file node is
+updated.  This is only used in case a file-watch update requires the insertion
+of a new file that, due to being recently created, does not have a git status
+cache entry.
 
 When OVERRIDE-STATUS is non-nil the FILE's cached git status will not be used.
 
-When EXCLUDE-PARENTS is non-nil only the given FILE only the file node is
-updated. This is only used in case a file-watch update requires the insertion of
-a new file that, due to being recently created, does not have a git status cache
-entry.
-
-FILE: Filepath
+FILE: FilePath
 EXCLUDE-PARENTS: Boolean
 OVERRIDE-STATUS: Boolean"
   (let* ((local-buffer (current-buffer))
@@ -310,7 +325,8 @@ OVERRIDE-STATUS: Boolean"
                    (when git-cache
                      (ht-set! git-cache file state))
                    (-when-let (pos (treemacs-find-visible-node file))
-                     (-let [face (treemacs--git-status-face state 'treemacs-git-unmodified-face)]
+                     (let* ((default (if (file-directory-p file) 'treemacs-directory-face 'treemacs-git-unmodified-face))
+                            (face (treemacs--git-status-face state default)))
                        (put-text-property
                         (treemacs-button-start pos) (treemacs-button-end pos)
                         'face face))))
@@ -326,22 +342,22 @@ OVERRIDE-STATUS: Boolean"
             (2 (ignore "No Change, Do Nothing"))
             (_
              (-let [err-str (treemacs--remove-trailing-newline (pfuture-output-from-buffer pfuture-buffer))]
-               (treemacs-log "Update of node \"%s\" failed with status \"%s\" and result"
+               (treemacs-log-err "Update of node \"%s\" failed with status \"%s\" and result"
                  file (treemacs--remove-trailing-newline status))
-               (treemacs-log "\"%s\"" (treemacs--remove-trailing-newline err-str))))))))))
+               (treemacs-log-err "\"%s\"" (treemacs--remove-trailing-newline err-str))))))))))
 
 (defun treemacs--collapsed-dirs-process (path project)
-  "Start a new process to determine dirs to collpase under PATH.
+  "Start a new process to determine directories to collapse under PATH.
 Only starts the process if PROJECT is locally accessible (i.e. exists, and
 is not remote.)
 Output format is an elisp list of string lists that's read directly.
 Every string list consists of the following elements:
  1) the extra text that must be appended in the view
- 2) The original full and uncollapsed path
+ 2) The original full and non-collapsed path
  3) a series of intermediate steps which are the result of appending the
     collapsed path elements onto the original, ending in
  4) the full path to the
-    directory that the collapsing leads to. For Example:
+    directory that the collapsing leads to.  For Example:
     (\"/26.0/elpa\"
      \"/home/a/Documents/git/treemacs/.cask\"
      \"/home/a/Documents/git/treemacs/.cask/26.0\"
@@ -354,13 +370,13 @@ Every string list consists of the following elements:
     (-let [default-directory path]
       (pfuture-new treemacs-python-executable
                    "-O"
-                   treemacs--dirs-to-collpase.py
+                   treemacs--dirs-to-collapse.py
                    path
                    (number-to-string treemacs-collapse-dirs)
                    (if treemacs-show-hidden-files "t" "x")))))
 
 (defun treemacs--parse-collapsed-dirs (future)
-  "Parse the output of collpsed dirs FUTURE.
+  "Parse the output of collapsed dirs FUTURE.
 Splits the output on newlines, splits every line on // and swallows the first
 newline."
   (when future
@@ -368,35 +384,77 @@ newline."
       (when (= 0 (process-exit-status future))
         (read output)))))
 
+(defun treemacs--prefetch-gitignore-cache (path)
+  "Pre-load all the git-ignored files in the given PATH.
+
+PATH is either the symbol `all', in which case the state of all projects in the
+current workspace is gathered instead, or a single project's path, when that
+project has just been added to the workspace.
+
+Required for `treemacs-hide-gitignored-files-mode' to properly work with
+deferred git-mode, as otherwise ignored files will not be hidden on the first
+run because the git cache has yet to be filled."
+  (if (eq path 'all)
+      (setf path (-map #'treemacs-project->path
+                       (treemacs-workspace->projects (treemacs-current-workspace))))
+    (setf path (list path)))
+  (pfuture-callback `(,treemacs-python-executable
+                      "-O"
+                      ,treemacs--find-ignored-files.py
+                      ,@path)
+    :on-error (ignore)
+    :on-success
+    (let ((ignore-pairs (read (pfuture-callback-output)))
+          (ignored-files nil))
+      (while ignore-pairs
+        (let* ((root  (pop ignore-pairs))
+               (file  (pop ignore-pairs))
+               (cache (ht-get treemacs--git-cache root)))
+          (unless cache
+            (setf cache (make-hash-table :size 20 :test 'equal))
+            (ht-set! treemacs--git-cache root cache))
+          (ht-set! cache file "!")
+          (push file ignored-files)))
+      (treemacs-run-in-every-buffer
+       (treemacs-save-position
+        (dolist (file ignored-files)
+          (when-let (treemacs-is-path-visible? file)
+            (treemacs-do-delete-single-node file))))))))
+
 (define-minor-mode treemacs-git-mode
   "Toggle `treemacs-git-mode'.
+
 When enabled treemacs will check files' git status and highlight them
-accordingly. This git integration is available in 3 variants: simple, extended
+accordingly.  This git integration is available in 3 variants: simple, extended
 and deferred.
 
 The simple variant will start a git status process whose output is parsed in
-elisp. This version is simpler and slightly faster, but incomplete - it will
+elisp.  This version is simpler and slightly faster, but incomplete - it will
 highlight only files, not directories.
 
 The extended variant requires a non-trivial amount of parsing to be done, which
-is achieved with python (specifically python3). It is slightly slower, but
-complete - both files and directories will be highlighted according to their
-git status.
+is achieved with python (specifically python3).  It is slightly slower, but
+complete - both files and directories will be highlighted according to their git
+status.
 
 The deferred variant is the same is extended, except the tasks of rendering
-nodes and highlighting them are separated. The former happens immediately, the
-latter after `treemacs-deferred-git-apply-delay' seconds of idle time. This may
-be faster (if not in truth then at least in appereance) as the git process is
-given a much greater amount of time to finish. The downside is that the effect
-of nodes changing their colors may be somewhat jarring, though this effect is
+nodes and highlighting them are separated.  The former happens immediately, the
+latter after `treemacs-deferred-git-apply-delay' seconds of idle time.  This may
+be faster (if not in truth then at least in appearance) as the git process is
+given a much greater amount of time to finish.  The downside is that the effect
+of nodes changing their colours may be somewhat jarring, though this issue is
 largely mitigated due to the use of a caching layer.
 
-All versions run asynchronously and are optimized for not doing more work than
+All versions run asynchronously and are optimised for not doing more work than
 is necessary, so their performance cost should, for the most part, be the
 constant time needed to fork a subprocess."
   :init-value nil
-  :global t
-  :lighter nil
+  :global     t
+  :lighter    nil
+  :group      'treemacs-git
+  ;; case when the mode is re-activated by `custom-set-minor-mode'
+  (when (and (eq t arg) treemacs--git-mode)
+    (setf arg treemacs--git-mode))
   (if treemacs-git-mode
       (if (memq arg '(simple extended deferred))
           (treemacs--setup-git-mode arg)
@@ -409,8 +467,8 @@ Use either ARG as git integration value of read it interactively."
   (interactive (list (-> (completing-read "Git Integration: " '("Simple" "Extended" "Deferred"))
                          (downcase)
                          (intern))))
-  (setq treemacs-git-mode arg)
-  (pcase treemacs-git-mode
+  (setf treemacs--git-mode arg)
+  (pcase treemacs--git-mode
     ((or 'extended 'deferred)
      (fset 'treemacs--git-status-process-function #'treemacs--git-status-process-extended)
      (fset 'treemacs--git-status-parse-function   #'treemacs--parse-git-status-extended))
@@ -423,12 +481,13 @@ Use either ARG as git integration value of read it interactively."
 
 (defun treemacs--tear-down-git-mode ()
   "Tear down `treemacs-git-mode'."
+  (setf treemacs--git-mode nil)
   (fset 'treemacs--git-status-process-function #'ignore)
   (fset 'treemacs--git-status-parse-function   (lambda (_) (ht))))
 
 (define-inline treemacs--get-or-parse-git-result (future)
   "Get the parsed git result of FUTURE.
-Parse and set it if it hasn't been done yet. If FUTURE is nil an empty hash
+Parse and set it if it hasn't been done yet.  If FUTURE is nil an empty hash
 table is returned.
 
 FUTURE: Pfuture process"
@@ -442,6 +501,43 @@ FUTURE: Pfuture process"
              result))
        (ht)))))
 
+(define-minor-mode treemacs-hide-gitignored-files-mode
+  "Toggle `treemacs-hide-gitignored-files-mode'.
+
+When enabled treemacs will hide files that are ignored by git.
+
+Some form of `treemacs-git-mode' *must* be enabled, otherwise this feature will
+have no effect.
+
+Both `extended' and `deferred' git-mode settings will work in full (in case of
+`deferred' git-mode treemacs will pre-load the list of ignored files so they
+will be hidden even on the first run).  The limitations of `simple' git-mode
+apply here as well: since it only knows about files and not directories only
+files will be hidden."
+  :init-value nil
+  :global     t
+  :lighter    nil
+  :group      'treemacs-git
+  (if treemacs-hide-gitignored-files-mode
+      (progn
+        (add-to-list 'treemacs-pre-file-insert-predicates
+                     #'treemacs-is-file-git-ignored?)
+        (when (and (eq 'deferred treemacs--git-mode)
+                   (not (get 'treemacs-hide-gitignored-files-mode
+                             :prefetch-done)))
+          (treemacs--prefetch-gitignore-cache 'all)
+          (put 'treemacs-hide-gitignored-files-mode :prefetch-done t)))
+    (setf treemacs-pre-file-insert-predicates
+          (delete #'treemacs-is-file-git-ignored?
+                  treemacs-pre-file-insert-predicates)))
+  (treemacs-run-in-every-buffer
+   (treemacs--do-refresh (current-buffer) 'all))
+  (when (called-interactively-p 'interactive)
+    (treemacs-pulse-on-success "Git-ignored files will now be %s"
+      (propertize
+       (if treemacs-hide-gitignored-files-mode "hidden." "displayed.")
+       'face 'font-lock-constant-face))) )
+
 (treemacs-only-during-init
  (let ((has-git    (not (null (executable-find "git"))))
        (has-python (not (null treemacs-python-executable))))
@@ -455,7 +551,7 @@ FUTURE: Pfuture process"
      (setf treemacs-collapse-dirs 3))
 
    (unless (or has-python (boundp 'treemacs-no-load-time-warnings))
-     (treemacs-log "Python3 not found, advanced git-mode and directory flattening features will be disabled."))))
+     (treemacs-log-failure "Python3 not found, advanced git-mode and directory flattening features will be disabled."))))
 
 (provide 'treemacs-async)
 
