@@ -1,6 +1,6 @@
 ;;; treemacs.el --- A tree style file viewer package -*- lexical-binding: t -*-
 
-;; Copyright (C) 2020 Alexander Miller
+;; Copyright (C) 2021 Alexander Miller
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -16,31 +16,38 @@
 ;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
-;;; Customize interface definitions.
+
+;; Customize interface definitions.
 
 ;;; Code:
 
-(require 'cl-lib)
+(require 's)
 (require 'widget)
 (require 'dash)
-(require 'f)
+
+(eval-when-compile
+  (require 'cl-lib))
 
 (defun treemacs--find-python3 ()
   "Determine the location of python 3."
   (--if-let (executable-find "python3") it
     (when (eq system-type 'windows-nt)
-      (->> "where python"
-           (shell-command-to-string)
-           (s-trim)
-           (s-lines)
-           (--first
-            (when (file-exists-p it)
-              (->> (concat (shell-quote-argument it) " --version")
-                   (shell-command-to-string)
-                   (s-trim)
-                   (s-replace "Python " "")
-                   (s-left 1)
-                   (version<= "3"))))))))
+      (condition-case _
+          (->> "where python"
+               (shell-command-to-string)
+               (s-trim)
+               (s-lines)
+               (--first
+                (when (file-exists-p it)
+                  (condition-case _
+                      (->> (concat (shell-quote-argument it) " --version")
+                           (shell-command-to-string)
+                           (s-trim)
+                           (s-replace "Python " "")
+                           (s-left 1)
+                           (version<= "3"))
+                    (error nil)))))
+        (error nil)))))
 
 (cl-macrolet
     ((define-action-widget (name include-default include-tab include-ret)
@@ -80,7 +87,7 @@
   :link '(url-link :tag "Repository" "https://github.com/Alexander-Miller/treemacs"))
 
 (defgroup treemacs-git nil
-  "Customizations for treemacs' git integration"
+  "Customisations for treemacs' git integration."
   :group 'treemacs
   :prefix "treemacs-"
   :link '(url-link :tag "Repository" "https://github.com/Alexander-Miller/treemacs"))
@@ -92,13 +99,13 @@
   :link '(url-link :tag "Repository" "https://github.com/Alexander-Miller/treemacs"))
 
 (defgroup treemacs-follow nil
-  "Customizations for the behaviour of the treemacs' file and tag following."
+  "Customisations for the behaviour of the treemacs' file and tag following."
   :group 'treemacs
   :prefix "treemacs-"
   :link '(url-link :tag "Repository" "https://github.com/Alexander-Miller/treemacs"))
 
 (defgroup treemacs-window nil
-  "Customizations for the behaviour of the treemacs window."
+  "Customisations for the behaviour of the treemacs window."
   :group 'treemacs
   :prefix "treemacs-"
   :link '(url-link :tag "Repository" "https://github.com/Alexander-Miller/treemacs"))
@@ -106,7 +113,7 @@
 (defcustom treemacs-indentation 2
   "The number of spaces or pixels each level is indented in the file tree.
 If the value is integer, indentation is created by repeating
-`treemacs-indentation-string'. If the value is a list of form '(INTEGER px),
+`treemacs-indentation-string'.  If the value is a list of form '(INTEGER px),
 indentation will be a space INTEGER pixels wide."
   :type '(choice (integer :tag "Spaces" :value 2)
                  (list :tag "Pixels"
@@ -114,6 +121,31 @@ indentation will be a space INTEGER pixels wide."
                        (const :tag "" px)))
   :group 'treemacs)
 
+(defcustom treemacs-litter-directories '("/node_modules" "/.venv" "/.cask")
+  "List of directories affected by `treemacs-cleanup-litter'.
+Every item in the list is a regular expression, to be recognised a directory
+must be matched with `string-match-p'.
+
+Regexp-quoting the items in this list is *not* necessary, the quoting will
+happen automatically when needed."
+  :type 'list
+  :group 'treemacs)
+
+(defcustom treemacs-read-string-input 'from-child-frame
+  "The function treemacs uses to read user input.
+Only applies to plaintext input, like when renaming a project, file or
+workspace.
+
+There are 2 options:
+ - `from-child-frame': will use the `cfrs' package to read input from a small
+   child frame pop-up.  Only available in GUI frames, otherwise the default
+   minibuffer input is used.
+ - `from-minibuffer': will read input from the minibuffer, same as baseline
+   Emacs."
+  :type '(choice (const :tag "With Child Frame Popup" 'from-child-frame)
+                 (const :tag "From the Minibuffer (Emacs Default)" 'from-minibuffer))
+  :group 'treemacs)
+
 (defcustom treemacs-move-forward-on-expand nil
   "When non-nil treemacs will move to the first child of an expanded node."
   :type 'boolean
@@ -125,16 +157,28 @@ Requires eldoc mode to be enabled."
   :type 'boolean
   :group 'treemacs)
 
+(defcustom treemacs-indent-guide-style 'line
+  "Determines the appearance of `treemacs-indent-guide-mode'.
+The choices are
+ - `line' for indent guides to use the ' ┃ ' character for every indentation
+   level
+ - `block' to use a thick '██' block interspersed at every second indentation
+   level"
+  :type '(choice (const :tag "Line" 'line) (const :tag "Block" 'block))
+  :group 'treemacs)
+
 (defcustom treemacs-indentation-string " "
   "The string that is for indentation in the file tree.
 Indentation is created by repeating this string `treemacs-indentation' many
-times. If `treemacs-indentation' is specified in pixels, this value is only used
-when there is no windowing system available."
+times.  If `treemacs-indentation' is specified in pixels, this value is only
+used when there is no windowing system available."
   :type 'string
   :group 'treemacs)
 
 (defcustom treemacs-show-hidden-files t
-  "Dotfiles will be shown if this is set to t and be hidden otherwise."
+  "Dotfiles will be shown if this is set to t and be hidden otherwise.
+
+Can be toggled by `treemacs-toggle-show-dotfiles'."
   :type 'boolean
   :group 'treemacs)
 
@@ -180,7 +224,7 @@ of how this config works and how to modify it."
 
 (defcustom treemacs-default-visit-action
   'treemacs-visit-node-no-split
-  "Defines the behavior of `treemacs-visit-node-default'."
+  "Defines the behaviour of `treemacs-visit-node-default'."
   :type 'treemacs-default-action
   :group 'treemacs)
 
@@ -197,8 +241,8 @@ of how this config works and how to modify it."
   "Defines the behaviour of `treemacs-RET-action'.
 
 Each alist element maps from a button state to the function that should be used
-for that state. The list of all possible button states is defined in
-`treemacs-valid-button-states'. Possible values are all treemacs-visit-node-*
+for that state.  The list of all possible button states is defined in
+`treemacs-valid-button-states'.  Possible values are all treemacs-visit-node-*
 functions as well as `treemacs-toggle-node' for simple open/close actions,
 though in general you can use any function that accepts the prefix arg as its
 single argument.
@@ -209,6 +253,23 @@ To keep the alist clean changes should not be made directly, but with
   :type '(alist :key-type symbol :value-type treemacs-ret-action)
   :group 'treemacs)
 
+(defcustom treemacs-COLLAPSE-actions-config
+  '((root-node-open   . treemacs-toggle-node)
+    (root-node-closed . treemacs-goto-parent-node)
+    (dir-node-open    . treemacs-toggle-node)
+    (dir-node-closed  . treemacs-goto-parent-node)
+    (file-node-open   . treemacs-toggle-node)
+    (file-node-closed . treemacs-goto-parent-node)
+    (tag-node-open    . treemacs-toggle-node)
+    (tag-node-closed  . treemacs-goto-parent-node)
+    (tag-node         . treemacs-goto-parent-node))
+  "Defines the behaviour of `treemacs-COLLAPSE-action'.
+
+See the doc string of `treemacs-RET-actions-config' for a detailed description
+of how this config works and how to modify it."
+  :type '(alist :key-type symbol :value-type treemacs-collapse-action)
+  :group 'treemacs)
+
 (defcustom treemacs-dotfiles-regex (rx bol "." (1+ any))
   "Files matching this regular expression count as dotfiles."
   :type 'regexp
@@ -219,33 +280,33 @@ To keep the alist clean changes should not be made directly, but with
 Files will still always be shown after directories.
 
 Valid values are:
- * alphabetic-asc,
- * alphabetic-desc,
- * alphabetic-case-insensitive-asc,
- * alphabetic-case-insensitive-desc,
- * size-asc,
- * size-desc,
- * mod-time-asc,
- * mod-time-desc
+ * `alphabetic-asc',
+ * `alphabetic-desc',
+ * `alphabetic-case-insensitive-asc',
+ * `alphabetic-case-insensitive-desc',
+ * `size-asc',
+ * `size-desc',
+ * `mod-time-asc',
+ * `mod-time-desc'
  * a custom function
 
 In the latter case it must be a function that can be passed to `sort' to sort
-absolute filepaths. For an example see `treemacs--sort-alphabetic-asc'
+absolute filepaths.  For an example see `treemacs--sort-alphabetic-asc'
 
 Note about performance:
-Treemacs does its best to optimize its performance critical path, it does so
+Treemacs does its best to optimise its performance critical path, it does so
 by doing as little work as possible and producing as little garbage as possible.
-Deciding on the order in which its nodes are inserted is a part of this path. As
-such certain tradeoffs need to be accounted far.
+Deciding on the order in which its nodes are inserted is a part of this path.
+As such certain trade-offs need to be accounted far.
 
-In plaintext: some sort settings are much slower than others. Alphabetic sorting
-\(the default) is fastest and causes no additional overhead (even when compared
-against foregoing sorting altogether).
+In plaintext: some sort settings are much slower than others.  Alphabetic
+sorting \(the default) is fastest and causes no additional overhead (even when
+compared against foregoing sorting altogether).
 
-Modification time sorting takes the middle, being ca. 4x slower than alphabetic.
-Sorting by size is slowest, being ca. 5-6x slower than alphabetic. It also
-produces the most garbage, making it more like for you to run into a garbage
-collection pause.
+Modification time sorting takes the middle, being ca.  4x slower than
+alphabetic.  Sorting by size is slowest, being ca.  5-6x slower than alphabetic.
+It also produces the most garbage, making it more like for you to run into a
+garbage collection pause.
 
 Lest these numbers scare you off keep in mind that they will likely have little
 to no effect on your usage of treemacs until you begin frequently refreshing
@@ -269,13 +330,13 @@ treemacs views containing hundreds or even thousands of nodes."
 Ignored files will *never* be shown in the treemacs buffer (unlike dotfiles)
 whose presence is controlled by `treemacs-show-hidden-files').
 
-Each predicate is a function that takes 2 arguments: a files's name and its
-absolute path and returns t if the file should be ignored and nil otherwise. A
+Each predicate is a function that takes 2 arguments: a file's name and its
+absolute path and returns t if the file should be ignored and nil otherwise.  A
 file which returns t for *any* function in this list counts as ignored.
 
 By default this list contains `treemacs--std-ignore-file-predicate' which
-filters out '.', '..', Emacs' lock files as well as flycheck's temp files, and
-therefore should not be directly overwritten, but added to and removed from
+filters out '.', '..', Emacs' lock files as well temp files created by flycheck,
+and therefore should not be directly overwritten, but added to and removed from
 instead.
 Additionally `treemacs--mac-ignore-file-predicate' is also included on
 Mac-derived operating systems (when `system-type' is `darwin')."
@@ -286,13 +347,15 @@ Mac-derived operating systems (when `system-type' is `darwin')."
   "List of predicates to test for files and directories that shouldn't be shown.
 The difference between this and `treemacs-ignored-file-predicates' is that the
 functions in this list will be called on files just before they would be
-rendered, when the files' git status information is now available. This for
-example allows to make files ignored by git invisible.
+rendered, when the files' git status information is now available.  This for
+example allows to make files ignored by git invisible (however this particular
+use-case is already covered by `treemacs-hide-gitignored-files-mode').
+
 The functions in this list are therefore expected to have a different signature:
-They must take two arguments - a file's absolute path and a hashtable that maps
-files to their git status. The files' paths are the table's keys, its values are
-characters (and not strings) indicating the file's git condition. The chars map
-map as follows: (the pattern is derived from 'git status --porcelain')
+They must take two arguments - a file's absolute path and a hash table that maps
+files to their git status.  The files' paths are the table's keys, its values
+are characters (and not strings) indicating the file's git condition.  The chars
+map map as follows: (the pattern is derived from 'git status --porcelain')
 
  * M - file is modified
  * U - file is in conflict
@@ -303,10 +366,7 @@ map as follows: (the pattern is derived from 'git status --porcelain')
 
 Otherwise the behaviour is the same as `treemacs-ignored-file-predicates', in
 that any one function returning t for a file means that this file will not
-be rendered.
-
-Since removing files ignored by git is the most likely use-case treemacs offers
-`treemacs-is-file-git-ignored?' to quickly make this possible."
+be rendered."
   :type 'list
   :group 'treemacs)
 
@@ -314,26 +374,26 @@ Since removing files ignored by git is the most likely use-case treemacs offers
   "How long (in milliseconds) to collect file events before refreshing.
 When treemacs receives a file change notification it doesn't immediately refresh
 and instead waits `treemacs-file-event-delay' milliseconds to collect further
-file change events. This is done so as to avoid refreshing multiple times in a
+file change events.  This is done so as to avoid refreshing multiple times in a
 short time.
 See also `treemacs-filewatch-mode'."
   :type 'integer
   :group 'treemacs)
 
 (defcustom treemacs-goto-tag-strategy 'refetch-index
-  "Inidicates how to move to a tag when its buffer is dead.
+  "Indicates how to move to a tag when its buffer is dead.
 The tags in the treemacs view store their position as markers (or overlays if
-semantic mode is on) pointing to a buffer. If that buffer is killed, or has
+semantic mode is on) pointing to a buffer.  If that buffer is killed, or has
 never really been open, as treemacs kills buffer after fetching their tags if
 they did no exist before, the stored positions become stale, and treemacs needs
-to use a different method to move to that tag. This variale sets that method.
+to use a different method to move to that tag.  This variable sets that method.
 
 Its possible values are:
 
  * refetch-index
    Call up the file's imenu index again and use its information to jump.
  * call-xref
-   Call `xref-find-definitions' to find the tag. Only available since emacs25.
+   Call `xref-find-definitions' to find the tag.
  * issue-warning
    Just issue a warning that the tag's position pointer is invalid."
   :type 'integer
@@ -347,9 +407,13 @@ directory.
 The value determines how many directories can be collapsed at once, both as a
 performance cap and to prevent too long directory names in the treemacs view.
 
-To minimize this option's impact on display performace the search for
+To minimise this option's impact on display performance the search for
 directories to collapse is done asynchronously in a python script and will thus
-only work when python installed. The script should work both on python2 and 3."
+only work when python installed.  The script should work both on python 2 and 3.
+
+If you experience incorrect display of CJK characters while using this feature
+you have to inform Emacs about your language environment using
+`set-language-environment'."
   :type 'integer
   :group 'treemacs)
 
@@ -374,6 +438,17 @@ The change will apply the next time a treemacs buffer is created."
   :type 'boolean
   :group 'treemacs)
 
+(defcustom treemacs-expand-after-init t
+  "When non-nil expand the first project after treemacs is first initialised.
+Might be superseded by `treemacs-follow-after-init'."
+  :type 'boolean
+  :group 'treemacs)
+
+(defcustom treemacs-expand-added-projects t
+  "When non-nil newly added projects will be expanded."
+  :type 'boolean
+  :group 'treemacs)
+
 (defcustom treemacs-recenter-after-project-jump 'always
   "Decides when to recenter view after moving between projects.
 Specifically applies to calling `treemacs-next-project' and
@@ -405,7 +480,7 @@ Possible values are:
 
 (defcustom treemacs-pulse-on-success t
   "When non-nil treemacs will pulse the current line as a success indicator.
-This applies to actions like `treemacs-copy-path-at-point'."
+This applies to actions like `treemacs-copy-relative-path-at-point'."
   :type 'boolean
   :group 'treemacs)
 
@@ -427,11 +502,11 @@ This value will apply when any one of the following options is set to
  * treemacs-recenter-after-project-expand
 
 In that case a call to `recenter' will be made when the distance between point
-and the top/bottom of the treemacs window is less then this many lines. The
+and the top/bottom of the treemacs window is less then this many lines.  The
 value is not an absolute line count, but a relative floating-point percentage,
 with 0.0 being 0% and 1.0 being 100%.
 This means that, for example, when this variable is set to 0.1 `recenter' will
-be called within a 10% distance of the treemacs window's top/bottom. For a
+be called within a 10% distance of the treemacs window's top/bottom.  For a
 window height of 40 lines that means point being within the first or last 4
 lines of the treemacs window.
 
@@ -481,18 +556,20 @@ Note that this does *not* take `scroll-margin' into account."
        2)))
   "The value for `imenu-generic-expression' treemacs uses in elisp buffers.
 More discriminating than the default as it distinguishes between functions,
-inline functions, macros, faces, variables, customizations and types."
+inline functions, macros, faces, variables, customisations and types.
+
+Can be set to nil to use the default value."
   :type 'alist
   :group 'treemacs)
 
 (defcustom treemacs-persist-file
-  (f-join user-emacs-directory ".cache" "treemacs-persist")
+  (expand-file-name ".cache/treemacs-persist" user-emacs-directory)
   "Path to the file treemacs uses to persist its state."
   :group 'treemacs
   :type 'string)
 
 (defcustom treemacs-last-error-persist-file
-  (f-join user-emacs-directory ".cache" "treemacs-persist-at-last-error")
+  (expand-file-name ".cache/treemacs-persist-at-last-error" user-emacs-directory)
   "File that stores the treemacs state as it was during the last load error."
   :group 'treemacs
   :type 'string)
@@ -519,39 +596,17 @@ missing project will not appear in the project list next time Emacs is started."
 
 (defcustom treemacs--fringe-indicator-bitmap
   (if (fboundp 'define-fringe-bitmap)
-      (define-fringe-bitmap 'treemacs--fringe-indicator-bitmap
-        (vector #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111
-                #b00000111))
+      (define-fringe-bitmap 'treemacs--fringe-indicator-bitmap-default (make-vector 200 #b00000111))
     'vertical-bar)
   "The fringe bitmap used by the fringe-indicator minor mode."
-  :type 'fringe-bitmap
-  :options (if (fboundp 'fringe-bitmaps)
-               (cons 'treemacs--fringe-indicator-bitmap fringe-bitmaps)
-             nil)
+  :type (append '(choice)
+                ;; :type is evaluated before the call to define-fringe-bitmap
+                ;; so 'treemacs--fringe-indicator-bitmap-default is not yet in
+                ;; fringe-bitmaps
+                '((const treemacs--fringe-indicator-bitmap-default))
+                ;; `fringe-bitmpas' is void in the CI build Emacs
+                (when (bound-and-true-p fringe-bitmaps)
+                  (mapcar (lambda (sym) `(const ,sym)) fringe-bitmaps)))
   :group 'treemacs)
 
 (defcustom treemacs-show-cursor nil
@@ -560,12 +615,12 @@ missing project will not appear in the project list next time Emacs is started."
   :group 'treemacs)
 
 (defcustom treemacs-directory-name-transformer #'identity
-  "Transformer function that is applied to directory names before rendering for any sort of cosmetic effect."
+  "Transformer to apply to directory names before rendering for cosmetic effect."
   :type 'function
   :group 'treemacs)
 
 (defcustom treemacs-file-name-transformer #'identity
-  "Transformer function that is applied to file names before rendering for any sort of cosmetic effect."
+  "Transformer to apply to file names before rendering for cosmetic effect."
   :type 'function
   :group 'treemacs)
 
@@ -574,9 +629,9 @@ missing project will not appear in the project list next time Emacs is started."
   "Minimum distance from the top/bottom for (tag-)follow mode to recenter.
 Treemacs will be calling `recenter' after following a file/tag if the distance
 between point and the top/bottom of the treemacs window is less then this many
-lines. The value is not an absolute line count, but a percentage, with 0.0
-being 0% and 1.0 being 100%. This means that when this variable is set to 0.1
-`recenter' will be called within a 10% distance of the window top/bottom. For a
+lines.  The value is not an absolute line count, but a percentage, with 0.0
+being 0% and 1.0 being 100%.  This means that when this variable is set to 0.1
+`recenter' will be called within a 10% distance of the window top/bottom.  For a
 window height of 40 lines that means point being within the first or last 4
 lines of the treemacs window.
 Will only take effect if `treemacs-recenter-after-tag-follow' and/or
@@ -587,10 +642,8 @@ Note that this does *not* take `scroll-margin' into account."
   :group 'treemacs-follow)
 
 (defcustom treemacs-follow-after-init nil
-  "When t always find and focus the current file when treemacs is built.
-
-A treemacs buffer is built when after calling `treemacs-init' or
-`treemacs-projectle-init'. This will ignore `treemacs-follow-mode'."
+  "When non-nil find the current file in treemacs after it is first initialised.
+Might supersede `treemacs-expand-after-init'."
   :type 'boolean
   :group 'treemacs-follow)
 
@@ -638,8 +691,9 @@ Possible values are:
 This means that treemacs will make sure that only the currently followed project
 is expanded while all others will remain collapsed.
 
-Setting this to t might lead to noticeable slowdowns, at least when `treemacs-git-mode'
-is enabled, since constantly expanding an entire project is fairly expensive."
+Setting this to t might lead to noticeable slowdowns, at least when
+`treemacs-git-mode' is enabled, since constantly expanding an entire project is
+fairly expensive."
   :type 'boolean
   :group 'treemacs-follow)
 
@@ -651,19 +705,19 @@ This is only relevant when using the deferred variant of git-mode."
 
 (defcustom treemacs-max-git-entries 5000
   "Maximum number of git status entries treemacs will process.
-Information for entries that number will be silently ignored. The 'entries'
-refer to the lines output by `git status --porcelain --ignored'. The limit does
+Information for entries that number will be silently ignored.  The 'entries'
+refer to the lines output by `git status --porcelain --ignored'.  The limit does
 not apply to the simple `treemacs-git-mode.'"
   :type 'number
   :group 'treemacs-git)
 
 (defcustom treemacs-python-executable (treemacs--find-python3)
   "The python executable used by treemacs.
-An asynchronous python process is used in two optional feaures:
+An asynchronous python process is used in two optional features:
 `treemacs-collapse-dirs' and the extended variant of `treemacs-git-mode'.
 
 There is generally only one reason to change this value: an extended
-`treemacs-git-mode' requires python3 to work. If the default python executable
+`treemacs-git-mode' requires python3 to work.  If the default python executable
 is pointing to python2 this config variable can be used to direct treemacs to
 the python3 binary."
   :type 'string
@@ -672,34 +726,59 @@ the python3 binary."
 (defcustom treemacs-git-command-pipe ""
   "Text to be appended to treemacs' git command.
 With `treemacs-git-mode' the command `git status --porcelain --ignored .' is
-run to fetch a directory's git information. The content of this variable will
-be appended to this git command. This might be useful in cases when git's output
-is so large that it leads to palpable delays, while setting
-`treemacs-max-git-entries' leads to loss of information. In such a scenario an
-additional filter statement (for example `| grep -v \"/vendor_dir/\"') can be used
-to reduce the size of the output to a manageable volume for treemacs."
+run to fetch a directory's git information.  The content of this variable will
+be appended to this git command.  This might be useful in cases when the output
+produced by git is so large that it leads to palpable delays, while setting
+`treemacs-max-git-entries' leads to loss of information.  In such a scenario an
+additional filter statement (for example `| grep -v \"/vendor_dir/\"') can be
+used to reduce the size of the output to a manageable volume for treemacs."
   :type 'string
   :group 'treemacs-git)
 
 (defcustom treemacs-is-never-other-window nil
   "When non-nil treemacs will use the `no-other-window' parameter.
 
-In practice means that treemacs will become invisible to commands like
+In practice it means that treemacs will become invisible to commands like
 `other-window' or `evil-window-left'."
   :type 'boolean
   :group 'treemacs-window)
 
+(defcustom treemacs-width-is-initially-locked t
+  "Indicates whether the width of the treemacs window is initially locked.
+A locked width means that changes it is only possible with the commands
+`treemacs-set-width' or `treemacs-toggle-fixed-width'."
+  :type 'boolean
+  :group 'treemacs-window)
+
+(defcustom treemacs-window-background-color nil
+  "Custom background colours for the treemacs window.
+Value must be a cons cell consisting of two colours: first the background of the
+treemacs window proper, then a second colour for treemacs' `hl-line' overlay
+marking the selected line."
+  :type '(cons color color)
+  :group 'treemacs-window)
+
 (defcustom treemacs-width 35
   "Width of the treemacs window."
   :type 'integer
   :group 'treemacs-window)
 
+(defcustom treemacs-wide-toggle-width 70
+  "When resizing, this value is added or substracted from the window width."
+  :type 'integer
+  :group 'treemacs-window)
+
+(defcustom treemacs-width-increment 1
+  "When resizing, this value is added or substracted from the window width."
+  :type 'integer
+  :group 'treemacs-window)
+
 (defcustom treemacs-display-in-side-window t
   "When non-nil treemacs will use a dedicated side-window.
 On the one hand this will alleviate issues of unequally sized window splits when
 treemacs is visible (since Emacs does not quite understand that treemacs has
-fixed window size). On the other hand it may lead to issues with other packages
-like shell-pop, as making treemacs a side-window makes it unsplittable."
+fixed window size).  On the other hand it may lead to issues with other packages
+like shell-pop, as making treemacs a side-window renders it un-splittable."
   :type 'boolean
   :group 'treemacs-window)
 
@@ -707,7 +786,7 @@ like shell-pop, as making treemacs a side-window makes it unsplittable."
   "When non-nil treemacs will have the `no-delete-other-windows' parameter.
 This parameter prevents the treemacs window from closing when calling
 `delete-other-windows' or when a command like `magit-status' would launch a new
-fullscreen buffer.
+full-screen buffer.
 Note that treemacs has its own delete-windows command with
 `treemacs-delete-other-windows' that behaves the same as `delete-other-windows',
 but won't close treemacs itself.
@@ -716,6 +795,31 @@ included in `treemacs-display-in-side-window'."
   :type 'boolean
   :group 'treemacs-window)
 
+(defcustom treemacs-user-header-line-format nil
+  "The header line used in the treemacs window.
+Can be set either to `treemacs-header-buttons-format' or any one of its
+constituent parts, or any other value acceptable for `header-line-format'."
+  :type 'string
+  :group 'treemacs-window)
+
+(defcustom treemacs-text-scale nil
+  "Optional scale for the text (not the icons) in the treemacs window.
+If set the value will be passed to `text-scale-increase'.  Both positive and
+negative values are possible."
+  :type 'integer
+  :group 'treemacs-window)
+
+(defcustom treemacs-select-when-already-in-treemacs 'move-back
+  "How `treemacs-select-window' behaves when treemacs is already selected.
+
+Possible values are:
+ - `stay' - remain in the treemacs windows, effectively doing nothing
+ - `move-back' - move point back to the most recently used window (as selected
+    by `get-mru-window')"
+  :type '(choice (const stay)
+                 (const move-back))
+  :group 'treemacs)
+
 (defcustom treemacs-position 'left
   "Position of treemacs buffer.
 
@@ -739,6 +843,37 @@ Will be called with the created file's or dir's path as the sole argument."
   :type 'hook
   :group 'treemacs-hooks)
 
+(defcustom treemacs-delete-file-functions nil
+  "Hooks to run whenever a file or directory is deleted.
+Applies only when using `treemacs-delete'.  Will be called with the created
+file's or dir's path as the sole argument."
+  :type 'hook
+  :group 'treemacs-hooks)
+
+(defcustom treemacs-rename-file-functions nil
+  "Hooks to run whenever a file or directory is renamed.
+
+Applies only when using `treemacs-rename'.  Will be called with 2 arguments: the
+file's old name, and the file's new name, both as absolute paths."
+  :type 'hook
+  :group 'treemacs-hooks)
+
+(defcustom treemacs-move-file-functions nil
+  "Hooks to run whenever a file or directory is moved.
+
+Applies only when using `treemacs-move-file'.  Will be called with 2 arguments:
+the file's old location, and the file's new location, both as absolute paths."
+  :type 'hook
+  :group 'treemacs-hooks)
+
+(defcustom treemacs-copy-file-functions nil
+  "Hooks to run whenever a file or directory is copied.
+
+Applies only when using `treemacs-copy-file'.  Will be called with 2 arguments:
+the original file's location, and the copy's location, both as absolute paths."
+  :type 'hook
+  :group 'treemacs-hooks)
+
 (defcustom treemacs-delete-project-functions nil
   "Hooks to run whenever a project is deleted.
 Will be called with the deleted project as the sole argument *after* it has been
@@ -748,7 +883,7 @@ deleted."
 
 (defcustom treemacs-rename-project-functions nil
   "Hooks to run whenever a project is renamed.
-Will be called with the renamed project and the old name as its argumens."
+Will be called with the renamed project and the old name as its arguments."
   :type 'hook
   :group 'treemacs-hooks)
 
@@ -767,7 +902,7 @@ been deleted."
 
 (defcustom treemacs-rename-workspace-functions nil
   "Hooks to run whenever a workspace is renamed.
-Will be called with the renamed workspace and the old name as its argumens."
+Will be called with the renamed workspace and the old name as its arguments."
   :type 'hook
   :group 'treemacs-hooks)
 
@@ -779,7 +914,7 @@ The current workspace will be available as `treemacs-current-workspace'."
 
 (defcustom treemacs-workspace-edit-hook nil
   "Hooks to run whenever the entire workspace layout has been rebuilt.
-This hook runs after `treemacs-finish-edit' has been called. After such an edit
+This hook runs after `treemacs-finish-edit' has been called.  After such an edit
 any number (including zero) of workspaces and projects may have been changed or
 created or deleted."
   :type 'hook
@@ -806,36 +941,36 @@ The following replacements are available:
 (defcustom treemacs-pre-refresh-hook nil
   "Hooks to run right before the refresh process for a project kicks off.
 During the refresh the project is effectively collapsed and then expanded again.
-This hook runs *before* that happens. It runs with treemacs as the
+This hook runs *before* that happens.  It runs with treemacs as the
 `current-buffer' and receives as its arguments all the information that treemacs
 collects for its refresh process:
  * The project being refreshed (might be 'all)
  * The current screen-line number (can be nil).
- * The current button. Might be nil if point is on the header line.
- * The current button's state. See also `treemacs-valid-button-states'. Is nil
+ * The current button.  Might be nil if point is on the header line.
+ * The current button's state.  See also `treemacs-valid-button-states'.  Is nil
    if the current button is nil.
- * The nearest file path, as collected with `treemacs--nearest-path'. Is nil if
+ * The nearest file path, as collected with `treemacs--nearest-path'.  Is nil if
    point is on the header.
- * The current button's tag path. Is nil if the current button is nil."
+ * The current button's tag path.  Is nil if the current button is nil."
   :type 'hook
   :group 'treemacs-hooks)
 
 (defcustom treemacs-post-refresh-hook nil
   "Hooks to run right before the refresh process is finished off.
 During the refresh the project is effectively collapsed and then expanded again.
-This hook runs *after* that has happened. It runs with treemacs as the
+This hook runs *after* that has happened.  It runs with treemacs as the
 `current-buffer' and receives as its arguments all the information that treemacs
-collects for its refresh process. Note that these values were collected at the
+collects for its refresh process.  Note that these values were collected at the
 start of the refresh, and may now be longer valid (for example the current
 button's position will be wrong, even if it wasn't deleted outright):
  * The project being refreshed (might be 'all)
  * The current screen-line number (can be nil).
- * The current button. Might be nil if point was on the header line.
- * The current button's state. See also `treemacs-valid-button-states'. Is nil
+ * The current button.  Might be nil if point was on the header line.
+ * The current button's state.  See also `treemacs-valid-button-states'.  Is nil
    if the current button is nil.
- * The nearest file path, as collected with `treemacs--nearest-path'. Is nil if
+ * The nearest file path, as collected with `treemacs--nearest-path'.  Is nil if
    point was on the header.
- * The current button's tag path. Is nil if the current button is nil."
+ * The current button's tag path.  Is nil if the current button is nil."
   :type 'hook
   :group 'treemacs-hooks)
 
@@ -851,10 +986,16 @@ The hooks will be run *after* the treemacs buffer was destroyed."
   :type 'hook
   :group 'treemacs-hooks)
 
-(defcustom treemacs-select-hook nil
+(define-obsolete-variable-alias 'treemacs-select-hook 'treemacs-select-functions "2.9")
+
+(defcustom treemacs-select-functions nil
   "Hooks to run when the treemacs window is selected.
-This only applies to commands like `treemacs' or `treemacs-select-window', not
-general window selection commands like `other-window'."
+The hook should accept one argument which is a symbol describing treemacs'
+visibility before the select was invoked, as it would have been returned by
+`treemacs-current-visibility'.
+
+This hook only applies to commands like `treemacs' or `treemacs-select-window',
+not general window selection commands like `other-window'."
   :type 'hook
   :group 'treemacs-hooks)
 
@@ -871,7 +1012,7 @@ current scope (frame or perspective) it was found for."
   "Decides how treemacs determines a file's extension.
 There are 2 options:
  - An extension should be everything past the *last* period of the file name.
-   In this case this shoud be set to `treemacs-last-period-regex-value'
+   In this case this should be set to `treemacs-last-period-regex-value'
  - An extension should be everything past the *first* period of the file name.
    In this case this should be set to `treemacs-first-period-regex-value'"
   :group 'treemacs
@@ -882,13 +1023,41 @@ There are 2 options:
   "Custom mode line format to be used in `treemacs-mode'.
 
 If nil treemacs will look for default value provided by `spaceline', `moody'
-or `doom-modeline' in that order. Finally, if none of these packages is
+or `doom-modeline' in that order.  Finally, if none of these packages is
 available \"Treemacs\" text will be displayed.
 
-For more specific information about formatting mode line check `mode-line-format'."
+Setting this to `none' will disable the modeline.
+
+For more specific information about formatting mode line check
+`mode-line-format'."
   :type 'sexp
   :group 'treemacs)
 
+(defcustom treemacs-workspace-switch-cleanup nil
+  "Indicates which, if any, buffers should be deleted on a workspace switch.
+Only applies when interactively calling `treemacs-switch-workspace'.
+
+Valid values are
+ - nil to do nothing
+ - `files' to delete buffers visiting files
+ - `all' to delete all buffers other than treemacs and the scratch buffer
+
+In any case treemacs itself and the scratch and messages buffer will be
+unaffected."
+  :type '(choice (const :tag "All Buffers" all)
+                 (const :tag "Only File Buffers" files)
+                 (const :tag "None" nil))
+  :group 'treemacs)
+
+(defcustom treemacs-imenu-scope 'everything
+  "Determines which items treemacs' imenu function will collect.
+There are 2 options:
+ - `everything' will collect entries from every project in the workspace.
+ - `current-project' will only gather the index for the project at point."
+  :type '(choice (const :tag "Everything" everything)
+                 (const :tag "Current Project Only" current-project))
+  :group 'treemacs)
+
 (provide 'treemacs-customization)
 
 ;;; treemacs-customization.el ends here
