@@ -1,6 +1,6 @@
 ;;; treemacs.el --- A tree style file viewer package -*- lexical-binding: t -*-
 
-;; Copyright (C) 2020 Alexander Miller
+;; Copyright (C) 2021 Alexander Miller
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -16,8 +16,10 @@
 ;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
-;;; Most of everything related to icons is handled here. Specifically the
-;;; definition, instantiation, customization, resizing and resetting of icons.
+
+;; Most of everything related to icons is handled here.  Specifically
+;; the definition, instantiation, customization, resizing and
+;; resetting of icons.
 
 ;;; Code:
 
@@ -26,7 +28,10 @@
 (require 's)
 (require 'ht)
 (require 'treemacs-themes)
-(eval-and-compile
+(require 'treemacs-logging)
+
+(eval-when-compile
+  (require 'cl-lib)
   (require 'inline)
   (require 'treemacs-macros))
 
@@ -36,7 +41,7 @@
 ;; gap wherever treemacs places an icon, regardess of transparency.
 ;; Using xpm instead of png images is one way to work around this, but it degrades icon
 ;; quality to an unacceptable degree. Another way is to directly change images' :background
-;; property. The backgrounds colors are derived from the current theme with `treemacs--setup-icon-highlight'
+;; property. The backgrounds colours are derived from the current theme with `treemacs--setup-icon-highlight'
 ;; and saved in `treemacs--selected-icon-background' and `treemacs--not-selected-icon-background'.
 ;; Every icon string stores two images with the proper :background values in its properties
 ;; 'img-selected and 'img-unselected. The 'display property of the icon in the current line
@@ -45,27 +50,25 @@
 ;; Since it is a marker in the treemacs buffer it is important for it to be reset whenever it might
 ;; become invalid.
 
-(defvar treemacs--not-selected-icon-background
-  (pcase (face-attribute 'default :background nil t)
-    ('unspecified
-     (prog1 "#2d2d31"
-       (unless (boundp 'treemacs-no-load-time-warnings)
-         (message "[Treemacs] Warning: coudn't find default background color for icons, falling back on #2d2d31."))))
-    ('unspecified-bg
-     (prog1 "#2d2d31"
-       (unless (boundp 'treemacs-no-load-time-warnings)
-         (message "[Treemacs] Warning: background color is unspecified, icons will likely look wrong. Falling back on #2d2d31."))))
-    (other other))
+(eval-and-compile
+  (defvar treemacs--not-selected-icon-background
+    (pcase (face-attribute 'default :background nil t)
+      ((or 'unspecified 'unspecified-bg "unspecified" "unspecified-bg")
+       (unless (or noninteractive (boundp 'treemacs-no-load-time-warnings))
+         (message "[Treemacs] Warning: coudn't find default background colour for icons, falling back on #2d2d31."))
+       "#2d2d31" )
+      (other other)))
   "Background for non-selected icons.")
 
-(defvar treemacs--selected-icon-background
-  (-let [bg (face-attribute 'hl-line :background nil t)]
-    (if (memq bg '(unspecified unspecified-b))
-        (prog1 treemacs--not-selected-icon-background
-          (unless (boundp 'treemacs-no-load-time-warnings)
-            (message "[Treemacs] Warning: couldn't find hl-line-mode's background color for icons, falling back on %s."
-                     treemacs--not-selected-icon-background)))
-      bg))
+(eval-and-compile
+  (defvar treemacs--selected-icon-background
+    (-let [bg (face-attribute 'hl-line :background nil t)]
+      (if (member bg '(unspecified unspecified-b "unspecified" "unspecified-bg"))
+          (prog1 treemacs--not-selected-icon-background
+            (unless (or noninteractive (boundp 'treemacs-no-load-time-warnings))
+              (message "[Treemacs] Warning: couldn't find hl-line-mode's background color for icons, falling back on %s."
+                       treemacs--not-selected-icon-background)))
+        bg)))
   "Background for selected icons.")
 
 (define-inline treemacs--set-img-property (image property value)
@@ -91,7 +94,7 @@
 (defmacro treemacs-get-icon-value (ext &optional tui theme)
   "Get the value of an icon for extension EXT.
 If TUI is non-nil the terminal fallback value is returned.
-THEME is the name of the theme to look in. Will cause an error if the theme
+THEME is the name of the theme to look in.  Will cause an error if the theme
 does not exist."
   `(let* ((theme ,(if theme
                       `(treemacs--find-theme ,theme)
@@ -101,18 +104,29 @@ does not exist."
                     `(treemacs-theme->gui-icons theme))))
      (ht-get icons ,ext)))
 
+(define-inline treemacs--get-local-face-background (face)
+  "Get the `:background' of the given face.
+Unlike `face-attribute' this will take the `faces-remapping-alist' into
+account."
+  (declare (side-effect-free t))
+  (inline-letevals (face)
+    (inline-quote
+     (--if-let (car (alist-get ,face face-remapping-alist))
+         (plist-get it :background)
+       (face-attribute ,face :background nil t)))))
+
 (defun treemacs--setup-icon-background-colors (&rest _)
   "Align icon backgrounds with current Emacs theme.
-Fetch the current Emacs theme's background & hl-line colors and inject them into
-the gui icons of every theme in `treemacs--themes'.
+Fetch the current Emacs theme's background & hl-line colours and inject them
+into the gui icons of every theme in `treemacs--themes'.
 Also called as advice after `load-theme', hence the ignored argument."
-  (let* ((default-background (face-attribute 'default :background nil t))
-         (hl-line-background (face-attribute 'hl-line :background nil t))
+  (let* ((default-background (treemacs--get-local-face-background 'default))
+         (hl-line-background (treemacs--get-local-face-background 'hl-line))
          (test-icon          (treemacs-get-icon-value 'dir-open))
          (icon-background    (treemacs--get-img-property (get-text-property 0 'img-unselected test-icon) :background))
          (icon-hl-background (treemacs--get-img-property (get-text-property 0 'img-selected test-icon) :background)))
     (when (memq default-background '(unspecified-bg unspecified))
-      (treemacs-log "Current theme fails to specify default background color, falling back on #2d2d31")
+      (treemacs-log-failure "Current theme fails to specify default background color, falling back on #2d2d31")
       (setq default-background "#2d2d31"))
     ;; make sure we only change all the icons' colors when we have to
     (unless (and (string= default-background icon-background)
@@ -132,7 +146,7 @@ Also called as advice after `load-theme', hence the ignored argument."
   "Will return non-nil when Emacs is unable to create images.
 In this scenario (usually caused by running Emacs without a graphical
 environment) treemacs will not create any of its icons and will be forced to
-permanently use its simple string icon fallack."
+permanently use its simple string icon fallback."
   (declare (pure t) (side-effect-free t))
   (inline-quote (not (image-type-available-p 'png))))
 
@@ -152,7 +166,7 @@ Aliased to the current theme's gui or tui icons.")
   "List of icons with variables.
 Every symbol S maps to a variable named \"treemacs-icons-S\". In addition S is
 also the key for the icon in both `treemacs-gui-icons' and `treemacs-tui-icons'.
-This combination alllows these icons-with-variables to be correctly changed in
+This combination allows these icons-with-variables to be correctly changed in
 `treemacs--select-icon-set'.")
 
 (defvar treemacs--icon-size 22
@@ -176,11 +190,11 @@ Necessary since root icons are not rectangular."
      (let ((height treemacs--icon-size)
            (width treemacs--icon-size))
        (when (and (integerp treemacs--icon-size)
-                  (s-ends-with? "root.png" ,file-path))
+                  (s-starts-with? "root-" ,file-path))
          (treemacs--root-icon-size-adjust width height))
        (if (and (integerp treemacs--icon-size) (image-type-available-p 'imagemagick))
            (create-image ,file-path 'imagemagick nil :ascent 'center :width width :height height)
-         (create-image ,file-path 'png nil :ascent 'center))))))
+         (create-image ,file-path 'png nil :ascent 'center :width width :height height))))))
 
 (define-inline treemacs--create-icon-strings (file fallback)
   "Create propertized icon strings for a given FILE image and TUI FALLBACK."
@@ -210,65 +224,84 @@ Necessary since root icons are not rectangular."
 (cl-defmacro treemacs-create-icon (&key file icon (fallback " ") icons-dir extensions)
   "Create an icon for the current theme.
 - FILE is a file path relative to the icon directory of the current theme.
-- ICON is a string of an already created icon. Mutually exclusive with FILE.
+- ICON is a string of an already created icon.  Mutually exclusive with FILE.
+- FALLBACK is the fallback string for situations where png images are
+  unavailable.  Can be set to `same-as-icon' to use the same value as ICON.
 - ICONS-DIR can optionally be used to overwrite the path used to find icons.
   Normally the current theme's icon-path is used, but it may be convenient to
   use another when calling `treemacs-modify-theme'.
-- FALLBACK is the fallback string for situations where png images are
-  unavailable.
 - EXTENSIONS is a list of file extensions the icon should be used for.
   Note that treemacs has a loose understanding of what constitutes an extension:
   it's either the text past the last period or the entire filename, so names
   like \".gitignore\" and \"Makefile\" can be matched as well.
-  An extension may also be a symbol instead of a string. In this case treemacs
+  An extension may also be a symbol instead of a string.  In this case treemacs
   will also create a variable named \"treemacs-icon-%s\" making it universally
   accessible."
   (treemacs-static-assert (or (null icon) (null file))
     "FILE and ICON arguments are mutually exclusive")
-  `(let* ((icons-dir ,(if icons-dir icons-dir `(treemacs-theme->path treemacs--current-theme)))
-          (icon-path ,(if file `(f-join icons-dir ,file) nil))
-          (icon-pair ,(if file `(treemacs--create-icon-strings icon-path ,fallback)
-                        `(cons ,(treemacs--splice-icon icon) ,fallback)))
-          (gui-icons (treemacs-theme->gui-icons treemacs--current-theme))
-          (tui-icons (treemacs-theme->tui-icons treemacs--current-theme))
-          (gui-icon  (car icon-pair))
-          (tui-icon  (cdr icon-pair)))
-     ,(unless file
-        `(progn
-           (ignore icon-path)
-           (ignore icons-dir)))
-     ;; prefer to have icons as empty strings with a display property for compatibility
-     ;; in e.g. dired, where an actual text icon would break `dired-goto-file-1'
-     (unless (get-text-property 0 'display gui-icon)
-       (setf gui-icon (propertize " " 'display gui-icon)))
-     ,@(->> (-filter #'symbolp extensions)
-            (--map `(progn (add-to-list 'treemacs--icon-symbols ',it)
-                           (defvar ,(intern (format "treemacs-icon-%s" it)) nil))))
-     (--each ',extensions
-       (ht-set! gui-icons it gui-icon)
-       (ht-set! tui-icons it tui-icon))))
+  (-let [ext-list (--map (if (stringp it) (downcase it) it)
+                         (if (symbolp extensions) (symbol-value extensions) extensions))]
+    `(let* ((fallback   ,(if (equal fallback (quote 'same-as-icon))
+                             icon
+                           fallback))
+            (icons-dir  ,(if icons-dir icons-dir `(treemacs-theme->path treemacs--current-theme)))
+            (icon-path  ,(if file `(treemacs-join-path icons-dir ,file) nil))
+            (icon-pair  ,(if file `(treemacs--create-icon-strings icon-path fallback)
+                           `(cons ,(treemacs--splice-icon icon) fallback)))
+            (gui-icons  (treemacs-theme->gui-icons treemacs--current-theme))
+            (tui-icons  (treemacs-theme->tui-icons treemacs--current-theme))
+            (gui-icon   (car icon-pair))
+            (tui-icon   (cdr icon-pair)))
+       ,(unless file
+          `(progn
+             (ignore icon-path)
+             (ignore icons-dir)))
+       ;; prefer to have icons as empty strings with a display property for compatibility
+       ;; in e.g. dired, where an actual text icon would break `dired-goto-file-1'
+       (unless (get-text-property 0 'display gui-icon)
+         (setf gui-icon (propertize " " 'display gui-icon)))
+       ,@(->> (-filter #'symbolp ext-list)
+              (--map `(progn (add-to-list 'treemacs--icon-symbols ',it)
+                             (defvar ,(intern (format "treemacs-icon-%s" it)) nil))))
+       (--each (quote ,ext-list)
+         (ht-set! gui-icons it gui-icon)
+         (ht-set! tui-icons it tui-icon)))))
 
 (treemacs-create-theme "Default"
-  :icon-directory (f-join treemacs-dir "icons/default")
+  :icon-directory (treemacs-join-path treemacs-dir "icons/default")
   :config
   (progn
     ;; directory and other icons
-    (treemacs-create-icon :file "vsc/root-closed.png" :extensions (root)       :fallback "")
-    (treemacs-create-icon :file "vsc/dir-closed.png"  :extensions (dir-closed) :fallback (propertize "+ " 'face 'treemacs-term-node-face))
-    (treemacs-create-icon :file "vsc/dir-open.png"    :extensions (dir-open)   :fallback (propertize "- " 'face 'treemacs-term-node-face))
-    (treemacs-create-icon :file "tags-leaf.png"       :extensions (tag-leaf)   :fallback (propertize "• " 'face 'font-lock-constant-face))
-    (treemacs-create-icon :file "tags-open.png"       :extensions (tag-open)   :fallback (propertize "▸ " 'face 'font-lock-string-face))
-    (treemacs-create-icon :file "tags-closed.png"     :extensions (tag-closed) :fallback (propertize "▾ " 'face 'font-lock-string-face))
-    (treemacs-create-icon :file "error.png"           :extensions (error)      :fallback (propertize "• " 'face 'font-lock-string-face))
-    (treemacs-create-icon :file "warning.png"         :extensions (warning)    :fallback (propertize "• " 'face 'font-lock-string-face))
-    (treemacs-create-icon :file "info.png"            :extensions (info)       :fallback (propertize "• " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "vsc/root-closed.png"   :extensions (root-closed) :fallback "")
+    (treemacs-create-icon :file "vsc/root-open.png"     :extensions (root-open)   :fallback "")
+    (treemacs-create-icon :file "vsc/dir-closed.png"    :extensions (dir-closed)  :fallback (propertize "+ " 'face 'treemacs-term-node-face))
+    (treemacs-create-icon :file "vsc/dir-open.png"      :extensions (dir-open)    :fallback (propertize "- " 'face 'treemacs-term-node-face))
+    (treemacs-create-icon :file "tags-leaf.png"         :extensions (tag-leaf)    :fallback (propertize "• " 'face 'font-lock-constant-face))
+    (treemacs-create-icon :file "tags-open.png"         :extensions (tag-open)    :fallback (propertize "▸ " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "tags-closed.png"       :extensions (tag-closed)  :fallback (propertize "▾ " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "error.png"             :extensions (error)       :fallback (propertize "• " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "warning.png"           :extensions (warning)     :fallback (propertize "• " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "info.png"              :extensions (info)        :fallback (propertize "• " 'face 'font-lock-string-face))
+    (treemacs-create-icon :file "mail.png"              :extensions (mail)        :fallback " ")
+    (treemacs-create-icon :file "bookmark.png"          :extensions (bookmark)    :fallback " ")
+    (treemacs-create-icon :file "svgrepo/screen.png"    :extensions (screen)      :fallback " ")
+    (treemacs-create-icon :file "svgrepo/house.png"     :extensions (house)       :fallback " ")
+    (treemacs-create-icon :file "svgrepo/list.png"      :extensions (list)        :fallback " ")
+    (treemacs-create-icon :file "svgrepo/repeat.png"    :extensions (repeat)      :fallback " ")
+    (treemacs-create-icon :file "svgrepo/suitcase.png"  :extensions (suitcase)    :fallback " ")
+    (treemacs-create-icon :file "svgrepo/close.png"     :extensions (close)       :fallback " ")
+    (treemacs-create-icon :file "svgrepo/cal.png"       :extensions (calendar)    :fallback " ")
+    (treemacs-create-icon :file "svgrepo/briefcase.png" :extensions (briefcase)   :fallback " ")
 
-    ;; ;; file icons
+    ;; file icons
     (treemacs-create-icon :file "txt.png"           :extensions (fallback))
     (treemacs-create-icon :file "emacs.png"         :extensions ("el" "elc"))
     (treemacs-create-icon :file "ledger.png"        :extensions ("ledger"))
     (treemacs-create-icon :file "yaml.png"          :extensions ("yml" "yaml" "travis.yml"))
-    (treemacs-create-icon :file "shell.png"         :extensions ("sh" "zsh" "fish"))
+    (treemacs-create-icon
+     :file "shell.png"
+     :extensions ("sh" "zsh" "zshrc" "zshenv" "fish" "zprofile" "zlogin" "zlogout" "bash"
+                  "bash_profile" "bashrc" "bash_login" "profile" "bash_aliases"))
     (treemacs-create-icon :file "pdf.png"           :extensions ("pdf"))
     (treemacs-create-icon :file "c.png"             :extensions ("c" "h"))
     (treemacs-create-icon :file "haskell.png"       :extensions ("hs" "lhs"))
@@ -279,7 +312,6 @@ Necessary since root icons are not rectangular."
     (treemacs-create-icon :file "asciidoc.png"      :extensions ("adoc" "asciidoc"))
     (treemacs-create-icon :file "rust.png"          :extensions ("rs"))
     (treemacs-create-icon :file "image.png"         :extensions ("jpg" "jpeg" "bmp" "svg" "png" "xpm" "gif"))
-    (treemacs-create-icon :file "emacs.png"         :extensions ("el" "elc"))
     (treemacs-create-icon :file "clojure.png"       :extensions ("clj" "cljs" "cljc"))
     (treemacs-create-icon :file "ts.png"            :extensions ("ts" "tsx"))
     (treemacs-create-icon :file "vue.png"           :extensions ("vue"))
@@ -291,8 +323,10 @@ Necessary since root icons are not rectangular."
     (treemacs-create-icon :file "jar.png"           :extensions ("jar"))
     (treemacs-create-icon :file "kotlin.png"        :extensions ("kt"))
     (treemacs-create-icon :file "scala.png"         :extensions ("scala"))
+    (treemacs-create-icon :file "gradle.png"        :extensions ("gradle"))
     (treemacs-create-icon :file "sbt.png"           :extensions ("sbt"))
     (treemacs-create-icon :file "go.png"            :extensions ("go"))
+    (treemacs-create-icon :file "systemd.png"       :extensions ("service" "timer"))
     (treemacs-create-icon :file "php.png"           :extensions ("php"))
     (treemacs-create-icon :file "js.png"            :extensions ("js" "jsx"))
     (treemacs-create-icon :file "babel.png"         :extensions ("babel"))
@@ -300,7 +334,7 @@ Necessary since root icons are not rectangular."
     (treemacs-create-icon :file "json.png"          :extensions ("json"))
     (treemacs-create-icon :file "julia.png"         :extensions ("jl"))
     (treemacs-create-icon :file "elx.png"           :extensions ("ex"))
-    (treemacs-create-icon :file "elx-light.png"     :extensions ("exs" "eex"))
+    (treemacs-create-icon :file "elx-light.png"     :extensions ("exs" "eex" "leex"))
     (treemacs-create-icon :file "ocaml.png"         :extensions ("ml" "mli" "merlin" "ocaml"))
     (treemacs-create-icon :file "direnv.png"        :extensions ("envrc"))
     (treemacs-create-icon :file "puppet.png"        :extensions ("pp"))
@@ -375,7 +409,7 @@ Necessary since root icons are not rectangular."
     (treemacs-create-icon :file "vsc/locale.png"    :extensions ("locale"))
     (treemacs-create-icon :file "vsc/manifest.png"  :extensions ("manifest"))
     (treemacs-create-icon :file "vsc/maven.png"     :extensions ("pom.xml" "maven.config" "extensions.xml" "settings.xml"))
-    (treemacs-create-icon :file "vsc/meson.png"     :extensions ("meson"))
+    (treemacs-create-icon :file "vsc/meson.png"     :extensions ("meson" "meson.build"))
     (treemacs-create-icon :file "vsc/nginx.png"     :extensions ("nginx.conf" "nginx"))
     (treemacs-create-icon :file "vsc/npm.png"       :extensions ("npmignore" "npmrc" "package.json" "package-lock.json" "npm-shrinwrap.json"))
     (treemacs-create-icon :file "vsc/wasm.png"      :extensions ("wasm" "wat"))
@@ -403,7 +437,7 @@ Necessary since root icons are not rectangular."
     (treemacs-create-icon :file "vsc/sql.png"       :extensions ("sql"))
     (treemacs-create-icon :file "vsc/toml.png"      :extensions ("toml"))
     (treemacs-create-icon :file "vsc/nim.png"       :extensions ("nim"))
-    (treemacs-create-icon :file "vsc/org.png"       :extensions ("org"))
+    (treemacs-create-icon :file "vsc/org.png"       :extensions ("org" "org_archive"))
     (treemacs-create-icon :file "vsc/perl.png"      :extensions ("pl" "pm" "perl"))
     (treemacs-create-icon :file "vsc/vim.png"       :extensions ("vimrc" "tridactylrc" "vimperatorrc" "ideavimrc" "vrapperrc"))
     (treemacs-create-icon :file "vsc/deps.png"      :extensions ("cask"))
@@ -417,11 +451,10 @@ Uses `treemacs-icon-fallback' as fallback."
   (inline-letevals (file)
     (inline-quote
      (let ((file-downcased (-> ,file (treemacs--filename) (downcase))))
-       (ht-get treemacs-icons
-               file-downcased
-               (ht-get treemacs-icons
-                       (treemacs--file-extension file-downcased)
-                       (with-no-warnings treemacs-icon-fallback)))))))
+       (or (ht-get treemacs-icons file-downcased)
+           (ht-get treemacs-icons
+                   (treemacs--file-extension file-downcased)
+                   (with-no-warnings treemacs-icon-fallback)))))))
 
 ;;;###autoload
 (defun treemacs-resize-icons (size)
@@ -433,25 +466,29 @@ If SIZE is 'nil' the icons are not resized and will retain their default size of
 There is only one size, the icons are square and the aspect ratio will be
 preserved when resizing them therefore width and height are the same.
 
-Resizing the icons only works if Emacs was built with ImageMagick support. If
-this is not the case this function will report an error.
+Resizing the icons only works if Emacs was built with ImageMagick support, or if
+using Emacs >= 27.1,which has native image resizing support.  If this is not the
+case this function will not have any effect.
 
 Custom icons are not taken into account, only the size of treemacs' own icons
 png are changed."
   (interactive "nIcon size in pixels: ")
-  (setq treemacs--icon-size size)
-  (treemacs--maphash (treemacs-theme->gui-icons treemacs--current-theme) (_ icon)
-    (let ((display        (get-text-property 0 'display icon))
-          (img-selected   (get-text-property 0 'img-selected icon))
-          (img-unselected (get-text-property 0 'img-unselected icon))
-          (width          treemacs--icon-size)
-          (height         treemacs--icon-size))
-      (when (eq 'image (car-safe display))
-        (when (s-ends-with? "root.png" (plist-get (cdr display) :file))
-          (treemacs--root-icon-size-adjust width height))
-        (dolist (property (list display img-selected img-unselected))
-          (plist-put (cdr property) :height height)
-          (plist-put (cdr property) :width width))))))
+  (if (not (or  (and (functionp 'image-transforms-p) (member 'scale (image-transforms-p)))
+                (image-type-available-p 'imagemagick)))
+      (treemacs-log-failure "Icons cannot be resized without image transforms or imagemagick support.")
+    (setq treemacs--icon-size size)
+    (treemacs--maphash (treemacs-theme->gui-icons treemacs--current-theme) (_ icon)
+      (let ((display        (get-text-property 0 'display icon))
+            (img-selected   (get-text-property 0 'img-selected icon))
+            (img-unselected (get-text-property 0 'img-unselected icon))
+            (width          treemacs--icon-size)
+            (height         treemacs--icon-size))
+        (when (eq 'image (car-safe display))
+          (when (s-ends-with? "root.png" (plist-get (cdr display) :file))
+            (treemacs--root-icon-size-adjust width height))
+          (dolist (property (list display img-selected img-unselected))
+            (plist-put (cdr property) :height height)
+            (plist-put (cdr property) :width width)))))))
 
 (defun treemacs--select-icon-set ()
   "Select the right set of icons for the current buffer.
@@ -470,17 +507,26 @@ TUI icons will be used if
             (value    (ht-get icons icon-symbol)))
         (set (make-local-variable variable) value)))))
 
+(define-inline treemacs-current-icons (&optional tui)
+  "Return the current theme's icons.
+Return the fallback icons if TUI is non-nil."
+  (inline-letevals (tui)
+    (inline-quote
+     (if ,tui
+         (treemacs-theme->tui-icons treemacs--current-theme)
+       (treemacs-theme->gui-icons treemacs--current-theme)))))
+
 ;;;###autoload
 (defun treemacs-define-custom-icon (icon &rest file-extensions)
   "Define a custom ICON for the current theme to use for FILE-EXTENSIONS.
 
 Note that treemacs has a very loose definition of what constitutes a file
 extension - it's either everything past the last period, or just the file's full
-name if there is no period. This makes it possible to match file names like
+name if there is no period.  This makes it possible to match file names like
 '.gitignore' and 'Makefile'.
 
 Additionally FILE-EXTENSIONS are also not case sensitive and will be stored in a
-downcased state."
+down-cased state."
   (unless icon
     (user-error "Custom icon cannot be nil"))
   (dolist (ext file-extensions)
@@ -488,13 +534,27 @@ downcased state."
              (downcase ext)
              (concat icon " "))))
 
+;;;###autoload
+(defun treemacs-define-custom-image-icon (file &rest file-extensions)
+  "Same as `treemacs-define-custom-icon' but for image icons instead of strings.
+FILE is the path to an icon image (and not the actual icon string).
+FILE-EXTENSIONS are all the (not case-sensitive) file extensions the icon
+should be used for."
+  (unless file
+    (user-error "Custom icon cannot be nil"))
+  (-let [icon (car (treemacs--create-icon-strings file " "))]
+    (dolist (ext file-extensions)
+      (ht-set! (treemacs-theme->gui-icons treemacs--current-theme)
+               (downcase ext)
+               icon))))
+
 ;;;###autoload
 (defun treemacs-map-icons-with-auto-mode-alist (extensions mode-icon-alist)
   "Remaps icons for EXTENSIONS according to `auto-mode-alist'.
 EXTENSIONS should be a list of file extensions such that they match the regex
 stored in `auto-mode-alist', for example '\(\".cc\"\).
 MODE-ICON-ALIST is an alist that maps which mode from `auto-mode-alist' should
-be assigned which treemacs icon, for exmaple
+be assigned which treemacs icon, for example
 '\(\(c-mode . treemacs-icon-c\)
   \(c++-mode . treemacs-icon-cpp\)\)"
   (dolist (extension extensions)
