@@ -1,6 +1,6 @@
 ;;; treemacs.el --- A tree style file viewer package -*- lexical-binding: t -*-
 
-;; Copyright (C) 2020 Alexander Miller
+;; Copyright (C) 2021 Alexander Miller
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -16,18 +16,19 @@
 ;; along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 ;;; Commentary:
-;;; General purpose macros, and those used in, but defined outside of
-;;; treemacs-core-utils.el are put here, to prevent using them before their
-;;; definition, hopefully preventing issues like #97.
+
+;; General purpose macros, and those used in, but defined outside of
+;; treemacs-core-utils.el are put here, to prevent using them before
+;; their definition, hopefully preventing issues like #97.
 
 ;;; Code:
 
 (require 'dash)
-(require 'f)
 (require 's)
 (require 'pcase)
-(require 'cl-lib)
+
 (eval-when-compile
+  (require 'cl-lib)
   (require 'gv))
 
 (declare-function treemacs--scope-store "treemacs-scope")
@@ -39,19 +40,10 @@ Creates a list of `declare-function' statements."
   (let ((imports (--map (list 'declare-function it file) functions)))
     `(progn ,@imports)))
 
-(defmacro treemacs-log (msg &rest args)
-  "Write a log statement given format string MSG and ARGS."
-  (declare (indent 1))
-  `(unless treemacs--no-messages
-     (message
-      "%s %s"
-      (propertize "[Treemacs]" 'face 'font-lock-keyword-face)
-      (format ,msg ,@args))))
-
 (defmacro treemacs-static-assert (predicate error-msg &rest error-args)
   "Assert for macros that triggers at expansion time.
 Tests PREDICATE and, if it evaluates to nil, throws an error with ERROR-MSG and
-ERROR-ARGS. Basically the same thing as `cl-assert', but does not (seem to)
+ERROR-ARGS.  Basically the same thing as `cl-assert', but does not (seem to)
 interfere with auto-completion."
   (declare (indent 1))
   `(unless ,predicate
@@ -60,22 +52,16 @@ interfere with auto-completion."
                    (list ,@error-args)))))
 
 (defmacro treemacs-with-writable-buffer (&rest body)
-  "Temporarily turn off read-ony mode to execute BODY."
+  "Temporarily turn off read-only mode to execute BODY."
   (declare (debug t))
   `(let (buffer-read-only)
      ,@body))
 
-(defmacro treemacs-without-messages (&rest body)
-  "Temporarily turn off messages to execute BODY."
-  (declare (debug t))
-  `(let ((treemacs--no-messages t))
-     ,@body))
-
 (defmacro treemacs-safe-button-get (button &rest properties)
   "Safely extract BUTTON's PROPERTIES.
 
 Using `button-get' on a button located in a buffer that is not the current
-buffer does not work, so this function will run the property extaction from
+buffer does not work, so this function will run the property extraction from
 inside BUTTON's buffer."
   `(with-current-buffer (marker-buffer ,button)
      ,(if (= 1 (length properties))
@@ -118,7 +104,8 @@ Log ERROR-MSG if no button is selected, otherwise run BODY."
      ,@body))
 
 (cl-defmacro treemacs-do-for-button-state
-    (&key on-root-node-open
+    (&key no-error
+          on-root-node-open
           on-root-node-closed
           on-file-node-open
           on-file-node-closed
@@ -127,12 +114,15 @@ Log ERROR-MSG if no button is selected, otherwise run BODY."
           on-tag-node-open
           on-tag-node-closed
           on-tag-node-leaf
-          on-nil
-          no-error)
+          on-nil)
   "Building block macro to execute a form based on the current node state.
-Will bind to current button to 'btn' for the executon of the action forms.
+Will bind to current button to 'btn' for the execution of the action forms.
 When NO-ERROR is non-nil no error will be thrown if no match for the button
-state is achieved."
+state is achieved.
+Otherwise either one of ON-ROOT-NODE-OPEN, ON-ROOT-NODE-CLOSED,
+ON-FILE-NODE-OPEN, ON-FILE-NODE-CLOSED, ON-DIR-NODE-OPEN, ON-DIR-NODE-CLOSED,
+ON-TAG-NODE-OPEN, ON-TAG-NODE-CLOSED, ON-TAG-NODE-LEAF or ON-NIL will be
+executed."
   (declare (debug (&rest [sexp form])))
   `(-if-let (btn (treemacs-current-button))
        (pcase (treemacs-button-get btn :state)
@@ -168,26 +158,33 @@ state is achieved."
      ,on-nil))
 
 (cl-defmacro treemacs--execute-button-action
-    (&key save-window
-          ensure-window-split
-          split-function
+    (&key no-match-explanation
           window
+          split-function
+          ensure-window-split
           dir-action
           file-action
           tag-section-action
           tag-action
-          no-match-explanation)
+          window-arg)
   "Infrastructure macro for setting up actions on different button states.
+
 Fetches the currently selected button and verifies it's in the correct state
 based on the given state actions.
+
 If it isn't it will log NO-MATCH-EXPLANATION, if it is it selects WINDOW (or
 `next-window' if none is given) and splits it with SPLIT-FUNCTION if given.
-DIR-ACTION, FILE-ACTION, TAG-SECTION-ACTION and TAG-ACTION are inserted into a
-`pcase' statement matching the buttons state. Project root nodes are treated the
-same common directory nodes.
-If ENSURE-WINDOW-SPLIT is t treemacs will vertically split the window if
+
+If ENSURE-WINDOW-SPLIT is non-nil treemacs will vertically split the window if
 treemacs is the only window to make sure a buffer is opened next to it, not
-under or below it."
+under or below it.
+
+DIR-ACTION, FILE-ACTION, TAG-SECTION-ACTION and TAG-ACTION are inserted into a
+`pcase' statement matching the buttons state.  Project root nodes are treated
+the same common directory nodes.
+
+WINDOW-ARG determines whether the treemacs windows should remain selected,
+\(single prefix arg), or deleted (double prefix arg)."
   (declare (debug (&rest [sexp form])))
   (let ((valid-states (list)))
     (when dir-action
@@ -238,9 +235,11 @@ under or below it."
                             (funcall visit-action btn)
                           (error "No match achieved even though button's state %s was part of the set of valid states %s"
                                  state ',valid-states))))
-                (when ,save-window
-                  (select-window current-window))))))))))
+                (pcase ,window-arg
+                  ('(4) (select-window current-window))
+                  ('(16) (delete-window current-window)))))))))))
 
+;; TODO(2021/08/28): RM
 (defmacro treemacs--without-filewatch (&rest body)
   "Run BODY without triggering the filewatch callback.
 Required for manual interactions with the file system (like deletion), otherwise
@@ -254,7 +253,7 @@ the on-delete code will run twice."
 Finally execute FINAL-FORM after the code to restore the position has run.
 
 This macro is meant for cases where a simple `save-excursion' will not do, like
-a refresh, which can potentially change the entire buffer layout. In pratice
+a refresh, which can potentially change the entire buffer layout.  In practice
 this means attempt first to keep point on the same file/tag, and if that does
 not work keep it on the same line."
   (declare (debug (form body)))
@@ -281,7 +280,7 @@ not work keep it on the same line."
          (treemacs-goto-file-node curr-file))
         ((or 'dir-node-open 'dir-node-closed 'file-node-open 'file-node-closed)
          ;; stay on the same file
-         (if (and (file-exists-p curr-file)
+         (if (and (treemacs-is-path-visible? curr-file)
                   (or treemacs-show-hidden-files
                       (not (s-matches? treemacs-dotfiles-regex (treemacs--filename curr-file)))))
              (treemacs-goto-file-node curr-file)
@@ -289,7 +288,7 @@ not work keep it on the same line."
            ;; try dodging to our immediate neighbours, if they are no longer visible either
            ;; keep going up
            (cl-labels
-               ((can-move-to (it) (and (file-exists-p it)
+               ((can-move-to (it) (and (treemacs-is-path-visible? it)
                                        (or treemacs-show-hidden-files
                                            (not (s-matches? treemacs-dotfiles-regex (treemacs--filename it)))))))
              (cond
@@ -319,13 +318,24 @@ not work keep it on the same line."
         (-let [buffer-point (point)]
           (with-selected-window curr-window
             ;; recenter starts counting at 0
-            (recenter (1- curr-win-line))
+            (-let [scroll-margin 0]
+              (recenter (1- curr-win-line)))
             (set-window-point (selected-window) buffer-point))))
       ,@final-form)))
 
+(defmacro treemacs-with-workspace (workspace &rest body)
+  "Use WORKSPACE as the current workspace when running BODY.
+Specifically this means that calls to `treemacs-current-workspace' will return
+WORKSPACE and if no workspace has been set for the current scope yet it will not
+be set either."
+  (declare (indent 1) (debug (form body)))
+  `(let ((treemacs-override-workspace ,workspace))
+     (ignore treemacs-override-workspace)
+     ,@body))
+
 (defmacro treemacs-run-in-every-buffer (&rest body)
   "Run BODY once locally in every treemacs buffer.
-Only includes treemacs filetree buffers, not extensions.
+Only includes treemacs file tree buffers, not extensions.
 Sets `treemacs-override-workspace' so calls to `treemacs-current-workspace'
 return the workspace of the active treemacs buffer."
   (declare (debug t))
@@ -333,43 +343,19 @@ return the workspace of the active treemacs buffer."
      (let ((buffer (treemacs-scope-shelf->buffer shelf))
            (workspace (treemacs-scope-shelf->workspace shelf)))
        (when (buffer-live-p buffer)
-         (-let [treemacs-override-workspace workspace]
-           (ignore treemacs-override-workspace)
+         (treemacs-with-workspace workspace
            (with-current-buffer buffer
              ,@body))))))
 
 (defmacro treemacs-run-in-all-derived-buffers (&rest body)
   "Run BODY once locally in every treemacs buffer.
-Inluceds *all* treemacs-mode-derived buffers, including extensions."
+Includes *all* treemacs-mode-derived buffers, including extensions."
   (declare (debug t))
   `(dolist (buffer (buffer-list))
      (when (buffer-local-value 'treemacs--in-this-buffer buffer)
        (with-current-buffer buffer
          ,@body))))
 
-(defmacro treemacs--defstruct (name &rest properties)
-  "Define a struct with NAME and PROPERTIES.
-Delegates to `cl-defstruct', creating a struct with a 'NAME->' `:conc-name' and
-foregoing typechecking for its properties for the hope of improved performance."
-  (declare (indent 1))
-  (-let [prefix (concat (symbol-name name) "->")]
-    `(progn
-       (cl-defstruct (,name (:conc-name ,(intern prefix)))
-         ,@properties)
-       ,@(--map
-          (let* ((prop-name (symbol-name (nth it properties)))
-                 (func-name (intern (concat prefix prop-name))))
-            `(progn
-               ;; ignore warnings since the accessors are defined twice
-               (with-no-warnings
-                 ;; redefine the accessors without the type checking
-                 (define-inline ,func-name (self)
-                   ,(format "Get the '%s' property of `%s' SELF." prop-name name)
-                   (declare (side-effect-free t))
-                   (inline-letevals (self)
-                     (inline-quote (aref ,',self ,(1+ it))))))))
-          (number-sequence 0 (1- (length properties)))))))
-
 (defmacro treemacs-only-during-init (&rest body)
   "Run BODY only when treemacs has not yet been loaded.
 Specifically only run it when (featurep 'treemacs) returns nil."
@@ -378,27 +364,27 @@ Specifically only run it when (featurep 'treemacs) returns nil."
      ,@body))
 
 (defmacro treemacs--maphash (table names &rest body)
-  "Iterate over entries of TABLE in BODY.
+  "Iterate over entries of TABLE with NAMES in BODY.
 Entry variables will bound based on NAMES which is a list of two elements."
   (declare (debug (sexp sexp body))
            (indent 2))
-  (let ((key-name (cl-first names))
-        (val-name (cl-second names)))
+  (let ((key-name (car names))
+        (val-name (cadr names)))
     `(maphash
       (lambda (,key-name ,val-name) ,@body)
       ,table)))
 
 (defmacro treemacs-error-return (error-msg &rest msg-args)
-  "Early return failure from `treemacs-block'.
-Will pass ERROR-MSG and MSG-ARGS to `treemacs-pulse-on-failure'."
+  "Interactive early return failure from `treemacs-block'.
+Will also pass ERROR-MSG and MSG-ARGS to `treemacs-pulse-on-failure'."
   (declare (indent 1) (debug (form body)))
   `(cl-return-from __body__
      (treemacs-pulse-on-failure ,error-msg ,@msg-args)))
 
 (defmacro treemacs-error-return-if (predicate error-msg &rest msg-args)
-  "Early return from `treemacs-block'.
-When PREDICATE returns non-nil value will pass ERROR-MSG and MSG-ARGS to
-`treemacs-pulse-on-failure'."
+  "Interactive early return from `treemacs-block'.
+Checks if PREDICATE returns a non-nil value, and will pass also ERROR-MSG and
+MSG-ARGS to `treemacs-pulse-on-failure'."
   (declare (indent 1) (debug (form sexp body)))
   `(when ,predicate
      (cl-return-from __body__
@@ -409,7 +395,7 @@ When PREDICATE returns non-nil value will pass ERROR-MSG and MSG-ARGS to
   (declare (debug t))
   `(cl-return-from __body__ ,ret))
 
-(defmacro treemacs-return-if (predicate ret)
+(defmacro treemacs-return-if (predicate &optional ret)
   "Early return from `treemacs-block'.
 When PREDICATE returns non-nil RET will be returned."
   (declare (indent 1) (debug (form sexp)))
@@ -446,23 +432,23 @@ This pattern is oftentimes used in treemacs, see also `treemacs-return-if',
 
 (defmacro treemacs-is-path (left op &optional right)
   "Readable utility macro for various path predicates.
-LEFT is a file path, RIGHT is either a path, project, or workspace while OP can
-take the following forms:
+LEFT is a file path, OP is the operator and RIGHT is either a path, project, or
+workspace.  OP can be one of the following:
 
  * `:same-as' will check for string equality
  * `:in' will check will check whether LEFT is a child or the same as RIGHT.
  * `:parent-of' will check whether LEFT is a parent of, and not equal to, RIGHT
  * `:in-project' will check whether LEFT is part of the project RIGHT
  * `:in-workspace' will check whether LEFT is part of the workspace RIGHT and
-   return the appropriate project when it is. If RIGHT is not given it will
+   return the appropriate project when it is.  If RIGHT is not given it will
    default to calling `treemacs-current-workspace'.
 
 LEFT and RIGHT are expected to be in treemacs canonical file path format (see
-also `treemacs--canonical-path').
+also `treemacs-canonical-path').
 
 Even if LEFT or RIGHT should be a form and not a variable it is guaranteed that
 they will be evaluated only once."
-  (declare (debug (form form form)))
+  (declare (debug (&rest form)))
   (treemacs-static-assert (memq op '(:same-as :in :parent-of :in-project :in-workspace))
     "Invalid treemacs-is-path operator: `%s'" op)
   (treemacs-static-assert (or (eq op :in-workspace) right)
@@ -521,7 +507,7 @@ treemacs buffer exists at all, BODY will be executed."
      ,@body))
 
 (defmacro treemacs-with-ignored-errors (ignored-errors &rest body)
-  "Evaluate BODY with specific errors ignored.
+  "Given list of specifically IGNORED-ERRORS evaluate BODY.
 
 IGNORED-ERRORS is a list of errors to ignore.  Each element is a list whose car
 is the error's type, and second item is a regex to match against error messages.
@@ -538,15 +524,25 @@ If any of the IGNORED-ERRORS matches, the error is suppressed and nil returned."
 
 (defmacro treemacs-debounce (guard delay &rest body)
   "Debounce a function call.
-Based on a timer GUARD variable run function BODY with the given DELAY."
+Based on a timer GUARD variable run function with the given DELAY and BODY."
   (declare (indent 2))
   `(unless ,guard
      (setf ,guard
            (run-with-idle-timer
             ,delay nil
             (lambda ()
-              ,@body
-              (setf ,guard nil))))))
+              (unwind-protect
+                  (progn ,@body)
+                (setf ,guard nil)))))))
+
+(defmacro treemacs-without-recenter (&rest body)
+  "Run BODY without the usual recentering for expanded nodes.
+Specifically `treemacs--no-recenter' will be set to 't' so that
+`treemacs--maybe-recenter' will have no effect during non-interactive updates
+triggered by e.g. filewatch-mode."
+  (declare (debug t))
+  `(let ((treemacs--no-recenter t))
+     ,@body))
 
 (provide 'treemacs-macros)
 
